from distutils.core import setup

setup(
	name = 'FLfPY',
	packages = ['FLfPY'],
    scripts = [],
	version = '0.0.0',
	description = 'FL tools',
	author = 'P. Y. Lin',
	author_email = 'pingyenlin@gmail.com',
    url = 'https://github.com/PingYenLin/ECGanalysis',
    keywords = [],
    classifiers = []
)